@@ -23,6 +23,9 @@
  *
 \* -------------------------------------------------------------------------- */
 #include <ruby.h>
+#if defined(USE_TBR) && defined(HAVE_THREAD_H)
+#include <ruby/thread.h>
+#endif
 #include <curl/curl.h>
 #include "membuffer.h"
 #include "sglib.h"  /* Simple Generic Library -> http://sglib.sourceforge.net */
@@ -35,6 +38,7 @@ static VALUE cSession = Qnil;
 static VALUE cRequest = Qnil;
 static VALUE ePatronError = Qnil;
 static VALUE eUnsupportedProtocol = Qnil;
+static VALUE eUnsupportedSSLVersion = Qnil;
 static VALUE eURLFormatError = Qnil;
 static VALUE eHostResolutionError = Qnil;
 static VALUE eConnectionFailed = Qnil;
@@ -73,19 +77,6 @@ static size_t session_write_handler(char* stream, size_t size, size_t nmemb, mem
   return size * nmemb;
 }
 
-static size_t session_read_handler(char* stream, size_t size, size_t nmemb, char **buffer) {
-  size_t result = 0;
-
-  if (buffer != NULL && *buffer != NULL) {
-    size_t len = size * nmemb;
-    char *s1 = strncpy(stream, *buffer, len);
-    result = strlen(s1);
-    *buffer += result;
-  }
-
-  return result;
-}
-
 /* A non-zero return value from the progress handler will terminate the current
  * request. We use this fact in order to interrupt any request when either the
  * user calls the "interrupt" method on the session or when the Ruby interpreter
@@ -109,7 +100,7 @@ struct curl_state_list {
   struct curl_state_list  *next;
 };
 
-#define CS_LIST_COMPARATOR(p, _state_) (p->state == _state_)
+#define CS_LIST_COMPARATOR(p, _state_) (p->state - _state_)
 
 static struct curl_state_list *cs_list = NULL;
 
@@ -128,9 +119,9 @@ static void cs_list_remove(struct curl_state *state) {
   struct curl_state_list *item = NULL;
 
   assert(state != NULL);
-  SGLIB_LIST_FIND_MEMBER(struct curl_state_list, cs_list, state, CS_LIST_COMPARATOR, next, item);
+
+  SGLIB_LIST_DELETE_IF_MEMBER(struct curl_state_list, cs_list, state, CS_LIST_COMPARATOR, next, item);
   if (item) {
-    SGLIB_LIST_DELETE(struct curl_state_list, cs_list, item, next);
     ruby_xfree(item);
   }
 }
@@ -203,10 +194,10 @@ static struct curl_state* get_curl_state(VALUE self) {
 /*----------------------------------------------------------------------------*/
 /* Method implementations                                                     */
 
-/* call-seq:
- *    Patron.libcurl_version   -> version string
- *
- * Returns the version of the embedded libcurl as a string.
+/*
+* Returns the version of the embedded libcurl.
+* 
+*  @return [String] libcurl version string
  */
 static VALUE libcurl_version(VALUE klass) {
   char* value = curl_version();
@@ -214,38 +205,42 @@ static VALUE libcurl_version(VALUE klass) {
   return rb_str_new2(value);
 }
 
-/* call-seq:
- *    Session.escape( string )   -> escaped string
+/*
+ * Escapes the provided string using libCURL URL escaping functions.
  *
- * URL escapes the provided string.
+ * @param [String] value plain string to URL-escape
+*  @return [String] the escaped string
  */
 static VALUE session_escape(VALUE self, VALUE value) {
-  struct curl_state *state = get_curl_state(self);
+  
   VALUE string = StringValue(value);
   char* escaped = NULL;
   VALUE retval = Qnil;
 
+  struct curl_state* state = curl_easy_init();
   escaped = curl_easy_escape(state->handle,
                              RSTRING_PTR(string),
                              (int) RSTRING_LEN(string));
 
   retval = rb_str_new2(escaped);
+  curl_easy_cleanup(state);
   curl_free(escaped);
 
   return retval;
 }
 
-/* call-seq:
- *    Session.unescape( string )   -> unescaped string
+/*
+ * Unescapes the provided string using libCURL URL escaping functions.
  *
- * Unescapes the provided string.
+ * @param [String] value URL-encoded String to unescape
+*  @return [String] unescaped (decoded) string
  */
 static VALUE session_unescape(VALUE self, VALUE value) {
-  struct curl_state *state = get_curl_state(self);
   VALUE string = StringValue(value);
   char* unescaped = NULL;
   VALUE retval = Qnil;
 
+  struct curl_state* state = curl_easy_init();
   unescaped = curl_easy_unescape(state->handle,
                                  RSTRING_PTR(string),
                                  (int) RSTRING_LEN(string),
@@ -253,6 +248,7 @@ static VALUE session_unescape(VALUE self, VALUE value) {
 
   retval = rb_str_new2(unescaped);
   curl_free(unescaped);
+  curl_easy_cleanup(state);
 
   return retval;
 }
@@ -260,9 +256,25 @@ static VALUE session_unescape(VALUE self, VALUE value) {
 /* Callback used to iterate over the HTTP headers and store them in an slist. */
 static int each_http_header(VALUE header_key, VALUE header_value, VALUE self) {
   struct curl_state *state = get_curl_state(self);
+  CURL* curl = state->handle;
+
   VALUE name = rb_obj_as_string(header_key);
   VALUE value = rb_obj_as_string(header_value);
   VALUE header_str = Qnil;
+  
+  // TODO: see how to combine this with automatic_content_encoding
+  if (rb_str_cmp(name, rb_str_new2("Accept-Encoding")) == 0) {
+    if (rb_funcall(value, rb_intern("include?"), 1, rb_str_new2("gzip"))) {
+      #ifdef CURLOPT_ACCEPT_ENCODING
+        curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "gzip");
+      #elif defined CURLOPT_ENCODING
+        curl_easy_setopt(curl, CURLOPT_ENCODING, "gzip");
+      #else
+        rb_raise(rb_eArgError,
+                "The libcurl version installed doesn't support 'gzip'.");
+      #endif
+    }
+  }
 
   header_str = rb_str_plus(name, rb_str_new2(": "));
   header_str = rb_str_plus(header_str, value);
@@ -294,6 +306,16 @@ static int formadd_files(VALUE data_key, VALUE data_value, VALUE self) {
   return 0;
 }
 
+// Set the given char pointer and it's length to be the CURL request body
+static void set_curl_request_body(CURL* curl, char* buf, curl_off_t len) {
+  curl_easy_setopt(curl, CURLOPT_POSTFIELDS, buf);
+  #ifdef CURLOPT_POSTFIELDSIZE_LARGE
+    curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE_LARGE, len);
+  #else
+    curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, len);
+  #endif
+}
+
 static void set_chunked_encoding(struct curl_state *state) {
   state->headers = curl_slist_append(state->headers, "Transfer-Encoding: chunked");
 }
@@ -325,10 +347,14 @@ static void set_options_from_request(VALUE self, VALUE request) {
   VALUE credentials           = Qnil;
   VALUE ignore_content_length = Qnil;
   VALUE insecure              = Qnil;
+  VALUE cacert                = Qnil;
+  VALUE ssl_version           = Qnil;
   VALUE buffer_size           = Qnil;
+  VALUE action_name           = rb_funcall(request, rb_intern("action"), 0);
+  VALUE a_c_encoding          = rb_funcall(request, rb_intern("automatic_content_encoding"), 0);
 
-  headers = rb_iv_get(request, "@headers");
-  if (!NIL_P(headers)) {
+  headers = rb_funcall(request, rb_intern("headers"), 0);
+  if (RTEST(headers)) {
     if (rb_type(headers) != T_HASH) {
       rb_raise(rb_eArgError, "Headers must be passed in a hash.");
     }
@@ -336,61 +362,59 @@ static void set_options_from_request(VALUE self, VALUE request) {
     rb_hash_foreach(headers, each_http_header, self);
   }
 
-  action = SYM2ID(rb_iv_get(request, "@action"));
+  action = SYM2ID(action_name);
+  if(rb_funcall(request, rb_intern("force_ipv4"), 0)) {
+    curl_easy_setopt(curl, CURLOPT_IPRESOLVE, CURL_IPRESOLVE_V4);
+  }
   if (action == rb_intern("get")) {
-    VALUE data = rb_iv_get(request, "@upload_data");
-    VALUE download_file = rb_iv_get(request, "@file_name");
+    VALUE data = rb_funcall(request, rb_intern("upload_data"), 0);
+    VALUE download_file = rb_funcall(request, rb_intern("file_name"), 0);
 
     curl_easy_setopt(curl, CURLOPT_HTTPGET, 1);
-    if (!NIL_P(data)) {
+    if (RTEST(data)) {
+      data = rb_funcall(data, rb_intern("to_s"), 0);
       long len = RSTRING_LEN(data);
       state->upload_buf = StringValuePtr(data);
-      curl_easy_setopt(curl, CURLOPT_UPLOAD, 1);
+      set_curl_request_body(curl, state->upload_buf, len);
       curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "GET");
-      curl_easy_setopt(curl, CURLOPT_READFUNCTION, &session_read_handler);
-      curl_easy_setopt(curl, CURLOPT_READDATA, &state->upload_buf);
-      curl_easy_setopt(curl, CURLOPT_INFILESIZE, len);
     }
-    if (!NIL_P(download_file)) {
-      state->download_file = open_file(download_file, "w");
+    if (RTEST(download_file)) {
+      state->download_file = open_file(download_file, "wb");
       curl_easy_setopt(curl, CURLOPT_WRITEDATA, state->download_file);
     } else {
       state->download_file = NULL;
     }
-  } else if (action == rb_intern("post") || action == rb_intern("put")) {
-    VALUE data = rb_iv_get(request, "@upload_data");
-    VALUE filename = rb_iv_get(request, "@file_name");
-    VALUE multipart = rb_iv_get(request, "@multipart");
-
-    if (!NIL_P(data) && NIL_P(multipart)) {
-      long len = RSTRING_LEN(data);
-
-      state->upload_buf = StringValuePtr(data);
+  } else if (action == rb_intern("post") || action == rb_intern("put") || action == rb_intern("patch")) {
+    VALUE data = rb_funcall(request, rb_intern("upload_data"), 0);
+    VALUE filename = rb_funcall(request, rb_intern("file_name"), 0);
+    VALUE multipart = rb_funcall(request, rb_intern("multipart"), 0);
 
+    if (action == rb_intern("post")) {
+      curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "POST");
+    } else if (action == rb_intern("put")) {
+      curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "PUT");
+    } else if (action == rb_intern("patch")) {
+      curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "PATCH");
+    }
+    
+    if (RTEST(data) && !RTEST(multipart)) {
+      data = rb_funcall(data, rb_intern("to_s"), 0);
       if (action == rb_intern("post")) {
         curl_easy_setopt(curl, CURLOPT_POST, 1);
-        curl_easy_setopt(curl, CURLOPT_POSTFIELDS, state->upload_buf);
-        curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, len);
-      } else {
-        curl_easy_setopt(curl, CURLOPT_UPLOAD, 1);
-        curl_easy_setopt(curl, CURLOPT_READFUNCTION, &session_read_handler);
-        curl_easy_setopt(curl, CURLOPT_READDATA, &state->upload_buf);
-        curl_easy_setopt(curl, CURLOPT_INFILESIZE, len);
       }
-    } else if (!NIL_P(filename) && NIL_P(multipart)) {
+      long len = RSTRING_LEN(data);
+      state->upload_buf = StringValuePtr(data);
+      set_curl_request_body(curl, state->upload_buf, len);
+    } else if (RTEST(filename) && !RTEST(multipart)) {
       set_chunked_encoding(state);
 
       curl_easy_setopt(curl, CURLOPT_UPLOAD, 1);
 
-      if (action == rb_intern("post")) {
-        curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "POST");
-      }
-
-      state->upload_file = open_file(filename, "r");
+      state->upload_file = open_file(filename, "rb");
       curl_easy_setopt(curl, CURLOPT_READDATA, state->upload_file);
-    } else if (!NIL_P(multipart)) {
+    } else if (RTEST(multipart)) {
       if (action == rb_intern("post")) {
-        if(!NIL_P(data) && !NIL_P(filename)) {
+        if(RTEST(data) && RTEST(filename)) {
           if (rb_type(data) == T_HASH && rb_type(filename) == T_HASH) {
             rb_hash_foreach(data, formadd_values, self);
             rb_hash_foreach(filename, formadd_files, self);
@@ -405,6 +429,19 @@ static void set_options_from_request(VALUE self, VALUE request) {
     } else {
       rb_raise(rb_eArgError, "Must provide either data or a filename when doing a PUT or POST");
     }
+
+  // support for data passed with a DELETE request (e.g.: used by elasticsearch)
+  } else if (action == rb_intern("delete")) {
+      VALUE data = rb_funcall(request, rb_intern("upload_data"), 0);
+      if (RTEST(data)) {
+        long len = RSTRING_LEN(data);
+        state->upload_buf = StringValuePtr(data);
+        curl_easy_setopt(curl, CURLOPT_POST, 1);
+        curl_easy_setopt(curl, CURLOPT_POSTFIELDS, state->upload_buf);
+        curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, len);
+      }
+      curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "DELETE");
+
   } else if (action == rb_intern("head")) {
     curl_easy_setopt(curl, CURLOPT_NOBODY, 1);
   } else {
@@ -415,59 +452,101 @@ static void set_options_from_request(VALUE self, VALUE request) {
   curl_easy_setopt(curl, CURLOPT_HTTPHEADER, state->headers);
   curl_easy_setopt(curl, CURLOPT_ERRORBUFFER, state->error_buf);
 
-  url = rb_iv_get(request, "@url");
-  if (NIL_P(url)) {
+  // Enable automatic content-encoding support via gzip/deflate if set in the request,
+  // see https://curl.haxx.se/libcurl/c/CURLOPT_ACCEPT_ENCODING.html
+  if(RTEST(a_c_encoding)) {
+    #ifdef CURLOPT_ACCEPT_ENCODING
+      curl_easy_setopt(curl, CURLOPT_ACCEPT_ENCODING, "");
+    #elif defined CURLOPT_ENCODING
+      curl_easy_setopt(curl, CURLOPT_ENCODING, "");
+    #else
+      rb_raise(rb_eArgError,
+        "The libcurl version installed doesn't support automatic content negotiation");
+    #endif
+  }
+  
+  url = rb_funcall(request, rb_intern("url"), 0);
+  if (!RTEST(url)) {
     rb_raise(rb_eArgError, "Must provide a URL");
   }
   curl_easy_setopt(curl, CURLOPT_URL, StringValuePtr(url));
-
-  timeout = rb_iv_get(request, "@timeout");
-  if (!NIL_P(timeout)) {
+  
+#ifdef CURLPROTO_HTTP
+  // Security: do not allow Curl to go looking on gopher/SMTP etc.
+  // Must prevent situations like this:
+  // https://hackerone.com/reports/115748
+  curl_easy_setopt(curl, CURLOPT_PROTOCOLS, CURLPROTO_HTTP | CURLPROTO_HTTPS);
+  curl_easy_setopt(curl, CURLOPT_REDIR_PROTOCOLS, CURLPROTO_HTTP | CURLPROTO_HTTPS);
+#endif
+    
+  timeout = rb_funcall(request, rb_intern("timeout"), 0);
+  if (RTEST(timeout)) {
     curl_easy_setopt(curl, CURLOPT_TIMEOUT, FIX2INT(timeout));
   }
 
-  timeout = rb_iv_get(request, "@connect_timeout");
-  if (!NIL_P(timeout)) {
+  timeout = rb_funcall(request, rb_intern("connect_timeout"), 0);
+  if (RTEST(timeout)) {
     curl_easy_setopt(curl, CURLOPT_CONNECTTIMEOUT, FIX2INT(timeout));
   }
 
-  redirects = rb_iv_get(request, "@max_redirects");
-  if (!NIL_P(redirects)) {
+  redirects = rb_funcall(request, rb_intern("max_redirects"), 0);
+  if (RTEST(redirects)) {
     int r = FIX2INT(redirects);
     curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, r == 0 ? 0 : 1);
     curl_easy_setopt(curl, CURLOPT_MAXREDIRS, r);
   }
 
-  proxy = rb_iv_get(request, "@proxy");
-  if (!NIL_P(proxy)) {
-      curl_easy_setopt(curl, CURLOPT_PROXY, StringValuePtr(proxy));
+  proxy = rb_funcall(request, rb_intern("proxy"), 0);
+  if (RTEST(proxy)) {
+    curl_easy_setopt(curl, CURLOPT_PROXY, StringValuePtr(proxy));
   }
 
-  proxy_type = rb_iv_get(request, "@proxy_type");
-  if (!NIL_P(proxy_type)) {
-    curl_easy_setopt(curl, CURLOPT_PROXYTYPE, FIX2INT(proxy_type));
+  proxy_type = rb_funcall(request, rb_intern("proxy_type"), 0);
+  if (RTEST(proxy_type)) {
+    curl_easy_setopt(curl, CURLOPT_PROXYTYPE, NUM2LONG(proxy_type));
   }
 
   credentials = rb_funcall(request, rb_intern("credentials"), 0);
-  if (!NIL_P(credentials)) {
-    curl_easy_setopt(curl, CURLOPT_HTTPAUTH, FIX2INT(rb_iv_get(request, "@auth_type")));
+  if (RTEST(credentials)) {
+    VALUE auth_type = rb_funcall(request, rb_intern("auth_type"), 0);
+    curl_easy_setopt(curl, CURLOPT_HTTPAUTH, NUM2LONG(auth_type));
     curl_easy_setopt(curl, CURLOPT_USERPWD, StringValuePtr(credentials));
   }
 
-  ignore_content_length = rb_iv_get(request, "@ignore_content_length");
-  if (!NIL_P(ignore_content_length)) {
+  ignore_content_length = rb_funcall(request, rb_intern("ignore_content_length"), 0);
+  if (RTEST(ignore_content_length)) {
     curl_easy_setopt(curl, CURLOPT_IGNORE_CONTENT_LENGTH, 1);
   }
 
-  insecure = rb_iv_get(request, "@insecure");
-  if(!NIL_P(insecure)) {
+  insecure = rb_funcall(request, rb_intern("insecure"), 0);
+  if(RTEST(insecure)) {
     curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, 0);
-    curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 1);
+    curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, 0);
   }
 
-  buffer_size = rb_iv_get(request, "@buffer_size");
-  if (!NIL_P(buffer_size)) {
-     curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, FIX2INT(buffer_size));
+  ssl_version = rb_funcall(request, rb_intern("ssl_version"), 0);
+  if(RTEST(ssl_version)) {
+    VALUE ssl_version_str = rb_funcall(ssl_version, rb_intern("to_s"), 0);
+    char* version = StringValuePtr(ssl_version_str);
+    if(strcmp(version, "SSLv2") == 0) {
+      curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_SSLv2);
+    } else if(strcmp(version, "SSLv3") == 0) {
+      curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_SSLv3);
+    } else if(strcmp(version, "TLSv1") == 0) {
+      curl_easy_setopt(curl, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1);
+    } else {
+      rb_raise(eUnsupportedSSLVersion, "Unsupported SSL version: %s", version);
+    }
+  }
+
+  cacert = rb_funcall(request, rb_intern("cacert"), 0);
+  if(RTEST(cacert)) {
+    curl_easy_setopt(curl, CURLOPT_CAINFO, StringValuePtr(cacert));
+  }
+
+  buffer_size = rb_funcall(request, rb_intern("buffer_size"), 0);
+  if (RTEST(buffer_size)) {
+     curl_easy_setopt(curl, CURLOPT_BUFFERSIZE, NUM2LONG(buffer_size));
   }
 
   if(state->debug_file) {
@@ -482,7 +561,8 @@ static VALUE create_response(VALUE self, CURL* curl, VALUE header_buffer, VALUE
   char* effective_url = NULL;
   long code = 0;
   long count = 0;
-
+  VALUE responseKlass = Qnil;
+  
   curl_easy_getinfo(curl, CURLINFO_EFFECTIVE_URL, &effective_url);
   args[0] = rb_str_new2(effective_url);
 
@@ -494,9 +574,10 @@ static VALUE create_response(VALUE self, CURL* curl, VALUE header_buffer, VALUE
 
   args[3] = header_buffer;
   args[4] = body_buffer;
-  args[5] = rb_iv_get(self, "@default_response_charset");
-
-  return rb_class_new_instance(6, args, rb_const_get(mPatron, rb_intern("Response")));
+  args[5] = rb_funcall(self, rb_intern("default_response_charset"), 0);
+  
+  responseKlass = rb_funcall(self, rb_intern("response_class"), 0);
+  return rb_class_new_instance(6, args, responseKlass);
 }
 
 /* Raise an exception based on the Curl error code. */
@@ -543,11 +624,18 @@ static VALUE perform_request(VALUE self) {
     curl_easy_setopt(curl, CURLOPT_WRITEDATA, body_buffer);
   }
 
-#if defined(HAVE_TBR) && defined(USE_TBR)
+#if (defined(HAVE_TBR) || defined(HAVE_TCWOGVL)) && defined(USE_TBR)
+#if defined(HAVE_TCWOGVL)
+  ret = (CURLcode) rb_thread_call_without_gvl(
+          (void *(*)(void *)) curl_easy_perform, curl,
+          RUBY_UBF_IO, 0
+        );
+#else
   ret = (CURLcode) rb_thread_blocking_region(
           (rb_blocking_function_t*) curl_easy_perform, curl,
           RUBY_UBF_IO, 0
         );
+#endif
 #else
   ret = curl_easy_perform(curl);
 #endif
@@ -556,7 +644,9 @@ static VALUE perform_request(VALUE self) {
     VALUE header_str = membuffer_to_rb_str(header_buffer);
     VALUE body_str = Qnil;
     if (!state->download_file) { body_str = membuffer_to_rb_str(body_buffer); }
-
+    
+    curl_easy_setopt(curl, CURLOPT_COOKIELIST, "FLUSH"); // Flush cookies to the cookie jar
+    
     return create_response(self, curl, header_str, body_str);
   } else {
     rb_raise(select_error(ret), "%s", state->error_buf);
@@ -585,14 +675,18 @@ static VALUE cleanup(VALUE self) {
     state->upload_file = NULL;
   }
 
+  if (state->post) {
+    curl_formfree(state->post);
+    state->post = NULL;
+    state->last = NULL;
+  }
+
   state->upload_buf = NULL;
 
   return Qnil;
 }
 
-/* call-seq:
- *    session.handle_request( request )   -> response
- *
+/*
  * Peform the actual HTTP request by calling libcurl. Each filed in the
  * +request+ object will be used to set the appropriate option on the libcurl
  * library. After the request completes, a Response object will be created and
@@ -601,17 +695,25 @@ static VALUE cleanup(VALUE self) {
  * In the event of an error in the libcurl library, a Ruby exception will be
  * created and raised. The exception will return the libcurl error code and
  * error message.
+ *
+ * @param request[Patron::Request] the request to use when filling the CURL options
+ * @return [Patron::Response] the result of calling `response_class` on the Session
  */
 static VALUE session_handle_request(VALUE self, VALUE request) {
   set_options_from_request(self, request);
   return rb_ensure(&perform_request, self, &cleanup, self);
 }
 
-/* call-seq:
- *    session.reset   -> session
- *
+/*
+ * FIXME: figure out how this method should be used at all given Session is not multithreaded.
+ * FIXME: also: what is the difference with `interrupt()` and also relationship with `cleanup()`?
  * Reset the underlying cURL session. This effectively closes all open
- * connections and disables debug output.
+ * connections and disables debug output. There is no need to call this method
+ * manually after performing a request, since cleanup is performed automatically
+ * but the method can be used from another thread
+ * to abort a request currently in progress.
+ *
+ * @return self
  */
 static VALUE session_reset(VALUE self) {
   struct curl_state *state;
@@ -627,11 +729,10 @@ static VALUE session_reset(VALUE self) {
   return self;
 }
 
-/* call-seq:
- *    session.interrupt   -> session
- *
- * Interrupt any currently executing request. This will cause the current
+/* Interrupt any currently executing request. This will cause the current
  * request to error and raise an exception.
+ *
+ * @return [void] This method always raises
  */
 static VALUE session_interrupt(VALUE self) {
   struct curl_state *state = get_curl_state(self);
@@ -639,18 +740,21 @@ static VALUE session_interrupt(VALUE self) {
   return self;
 }
 
-/* call-seq:
- *    session.enable_cookie_session( file )   -> session
- *
+/*
  * Turn on cookie handling for this session, storing them in memory by
- * default or in +file+ if specified. The +file+ must be readable and
+ * default or in +file+ if specified. The `file` must be readable and
  * writable. Calling multiple times will add more files.
+ * FIXME: what does the empty string actually do here?
+* 
+ * @param [String] file path to the existing cookie file, or nil to store in memory.
+*  @return self
  */
-static VALUE enable_cookie_session(VALUE self, VALUE file) {
+static VALUE add_cookie_file(VALUE self, VALUE file) {
   struct curl_state *state = get_curl_state(self);
   CURL* curl = state->handle;
   char* file_path = NULL;
 
+  // FIXME: http://websystemsengineering.blogspot.nl/2013/03/curloptcookiefile-vs-curloptcookiejar.html
   file_path = RSTRING_PTR(file);
   if (file_path != NULL && strlen(file_path) != 0) {
     curl_easy_setopt(curl, CURLOPT_COOKIEJAR, file_path);
@@ -660,10 +764,11 @@ static VALUE enable_cookie_session(VALUE self, VALUE file) {
   return self;
 }
 
-/* call-seq:
- *    session.set_debug_file( file )   -> session
- *
+/*
  * Enable debug output to stderr or to specified +file+.
+ *
+ * @param [String, nil] file path to the debug file, or nil to write to STDERR
+*  @return self
  */
 static VALUE set_debug_file(VALUE self, VALUE file) {
   struct curl_state *state = get_curl_state(self);
@@ -672,7 +777,7 @@ static VALUE set_debug_file(VALUE self, VALUE file) {
   session_close_debug_file(state);
 
   if(file_path != NULL && strlen(file_path) != 0) {
-    state->debug_file = open_file(file, "w");
+    state->debug_file = open_file(file, "wb");
   } else {
     state->debug_file = stderr;
   }
@@ -695,6 +800,7 @@ void Init_session_ext() {
   ePatronError = rb_const_get(mPatron, rb_intern("Error"));
 
   eUnsupportedProtocol = rb_const_get(mPatron, rb_intern("UnsupportedProtocol"));
+  eUnsupportedSSLVersion = rb_const_get(mPatron, rb_intern("UnsupportedSSLVersion"));
   eURLFormatError = rb_const_get(mPatron, rb_intern("URLFormatError"));
   eHostResolutionError = rb_const_get(mPatron, rb_intern("HostResolutionError"));
   eConnectionFailed = rb_const_get(mPatron, rb_intern("ConnectionFailed"));
@@ -708,26 +814,31 @@ void Init_session_ext() {
   cRequest = rb_define_class_under(mPatron, "Request", rb_cObject);
   rb_define_alloc_func(cSession, session_alloc);
 
+  // Make "escape" available both as a class method and as an instance method,
+  // to make it usable to the Util module which does not have access to
+  // the Session object internally
+  rb_define_singleton_method(cSession, "escape",   session_escape,         1);
   rb_define_method(cSession, "escape",         session_escape,         1);
+  rb_define_singleton_method(cSession, "unescape",   session_unescape,         1);
   rb_define_method(cSession, "unescape",       session_unescape,       1);
+
   rb_define_method(cSession, "handle_request", session_handle_request, 1);
   rb_define_method(cSession, "reset",          session_reset,          0);
   rb_define_method(cSession, "interrupt",      session_interrupt,      0);
-  rb_define_method(cSession, "enable_cookie_session", enable_cookie_session, 1);
+  rb_define_method(cSession, "add_cookie_file", add_cookie_file, 1);
   rb_define_method(cSession, "set_debug_file", set_debug_file, 1);
   rb_define_alias(cSession, "urlencode", "escape");
   rb_define_alias(cSession, "urldecode", "unescape");
 
-  rb_define_const(cRequest, "AuthBasic",  INT2FIX(CURLAUTH_BASIC));
-  rb_define_const(cRequest, "AuthDigest", INT2FIX(CURLAUTH_DIGEST));
-  rb_define_const(cRequest, "AuthAny",    INT2FIX(CURLAUTH_ANY));
+  rb_define_const(cRequest, "AuthBasic",  LONG2NUM(CURLAUTH_BASIC));
+  rb_define_const(cRequest, "AuthDigest", LONG2NUM(CURLAUTH_DIGEST));
+  rb_define_const(cRequest, "AuthAny",    LONG2NUM(CURLAUTH_ANY));
 
   mProxyType = rb_define_module_under(mPatron, "ProxyType");
-  rb_define_const(mProxyType, "HTTP", INT2FIX(CURLPROXY_HTTP));
-  rb_define_const(mProxyType, "HTTP_1_0", INT2FIX(CURLPROXY_HTTP_1_0));
-  rb_define_const(mProxyType, "SOCKS4", INT2FIX(CURLPROXY_SOCKS4));
-  rb_define_const(mProxyType, "SOCKS5", INT2FIX(CURLPROXY_SOCKS5));
-  rb_define_const(mProxyType, "SOCKS4A", INT2FIX(CURLPROXY_SOCKS4A));
-  rb_define_const(mProxyType, "SOCKS5_HOSTNAME", INT2FIX(CURLPROXY_SOCKS5_HOSTNAME));
+  rb_define_const(mProxyType, "HTTP", LONG2NUM(CURLPROXY_HTTP));
+  rb_define_const(mProxyType, "HTTP_1_0", LONG2NUM(CURLPROXY_HTTP_1_0));
+  rb_define_const(mProxyType, "SOCKS4", LONG2NUM(CURLPROXY_SOCKS4));
+  rb_define_const(mProxyType, "SOCKS5", LONG2NUM(CURLPROXY_SOCKS5));
+  rb_define_const(mProxyType, "SOCKS4A", LONG2NUM(CURLPROXY_SOCKS4A));
+  rb_define_const(mProxyType, "SOCKS5_HOSTNAME", LONG2NUM(CURLPROXY_SOCKS5_HOSTNAME));
 }
-
