@@ -21,6 +21,8 @@
 ## THE SOFTWARE.
 ##
 ## -------------------------------------------------------------------
+
+
 require File.expand_path("./spec") + '/spec_helper.rb'
 
 
@@ -34,12 +36,12 @@ describe Patron::Request do
 
     it "should accept :get, :put, :post, :delete and :head" do
       [:get, :put, :post, :delete, :head, :copy].each do |action|
-        lambda {@request.action = action}.should_not raise_error
+        expect {@request.action = action}.not_to raise_error
       end
     end
 
     it "should raise an exception when assigned a bad value" do
-      lambda {@request.action = :foo}.should raise_error(ArgumentError)
+      expect {@request.action = :foo}.to raise_error(ArgumentError)
     end
 
   end
@@ -47,11 +49,11 @@ describe Patron::Request do
   describe :timeout do
 
     it "should raise an exception when assigned a negative number" do
-      lambda {@request.timeout = -1}.should raise_error(ArgumentError)
+      expect {@request.timeout = -1}.to raise_error(ArgumentError)
     end
 
     it "should raise an exception when assigned 0" do
-      lambda {@request.timeout = 0}.should raise_error(ArgumentError)
+      expect {@request.timeout = 0}.to raise_error(ArgumentError)
     end
 
   end
@@ -59,7 +61,7 @@ describe Patron::Request do
   describe :max_redirects do
 
     it "should raise an error when assigned an integer smaller than -1" do
-      lambda {@request.max_redirects = -2}.should raise_error(ArgumentError)
+      expect {@request.max_redirects = -2}.to raise_error(ArgumentError)
     end
 
   end
@@ -67,7 +69,7 @@ describe Patron::Request do
   describe :headers do
 
     it "should raise an error when assigned something other than a hash" do
-      lambda {@request.headers = :foo}.should raise_error(ArgumentError)
+      expect {@request.headers = :foo}.to raise_error(ArgumentError)
     end
 
   end
@@ -75,11 +77,11 @@ describe Patron::Request do
   describe :buffer_size do
 
     it "should raise an exception when assigned a negative number" do
-      lambda {@request.buffer_size = -1}.should raise_error(ArgumentError)
+      expect {@request.buffer_size = -1}.to raise_error(ArgumentError)
     end
 
     it "should raise an exception when assigned 0" do
-      lambda {@request.buffer_size = 0}.should raise_error(ArgumentError)
+      expect {@request.buffer_size = 0}.to raise_error(ArgumentError)
     end
 
   end
@@ -87,18 +89,18 @@ describe Patron::Request do
   describe :eql? do
 
     it "should return true when two requests are equal" do
-      @request.should eql(Patron::Request.new)
+      expect(@request).to eql(Patron::Request.new)
     end
 
     it "should return false when two requests are not equal" do
       req = Patron::Request.new
       req.action = :post
-      @request.should_not eql(req)
+      expect(@request).not_to eql(req)
     end
 
   end
 
   it "should be able to serialize and deserialize itself" do
-    Marshal.load(Marshal.dump(@request)).should eql(@request)
+    expect(Marshal.load(Marshal.dump(@request))).to eql(@request)
   end
 end
